#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sched.h>
#include "pz.h"
#include "util.h"

typedef struct {
    int x, y;
} coord_t;

#ifdef IPOD
#define GNUGO_COMMAND "/home/gnugo"
#else
#define GNUGO_COMMAND "/usr/games/bin/gnugo"
#endif

#define GNUGO_ARGS_9 "-M 0.1","--never-resign","--boardsize=9","--level=0","--quiet","--mode=gtp","--gtp-input=/tmp/pipe_to_gnugo", (char *)NULL
#define GNUGO_ARGS_13 "-M 0.1","--never-resign","--boardsize=13","--level=0","--quiet","--mode=gtp","--gtp-input=/tmp/pipe_to_gnugo", (char *)NULL

char gnugo_pipe_to[] = "/tmp/pipe_to_gnugo";
char gnugo_pipe_from[] = "/tmp/pipe_from_gnugo";
FILE *gnugo_stdin; 
FILE *gnugo_stdout;
pid_t gnugo_pid;

// return values from gnugo_poll
enum {GP_NOTHING, GP_MOVEOK, GP_INVALID, GP_PASS};

// Logging
int logfile = 0;

// Game settings
int board_size = 9; // 9, 13 currently supported
int tilesize = 12;
float komi = 0;
int handi = 0;
int b_computer = 1, w_computer = 0;

// Game state
enum {MAINMENU, SETUP, GAMEMENU, WHITE_TURN, WHITE_WAIT,
      BLACK_TURN, BLACK_WAIT, COUNT_SCORE, GAMEOVER, 
      EXITGAME, QUIT} gamestate = MAINMENU;
coord_t w_last, b_last;
int b_passes, w_passes;
int b_captures, w_captures;
char winner;
float score;
int movenum;

//
PzModule *module;
PzWidget *widget;
ttk_surface tiles_9;
ttk_surface tiles_13;
ttk_surface backbuf;
ttk_font font;
coord_t cursor_pos;
coord_t cpos_old;

ttk_color col_black, col_white, col_green;

// Tiles and board
enum {T_BLACK, T_WHITE, T_EMPTY, T_A, T_RIGHT, T_LEFT, T_TOP, T_BOTTOM,
      T_TR, T_BR, T_TL, T_BL, N_TILES};
enum {RD_FULL, RD_CURSOR} redraw = RD_FULL;
enum {HORIZ, VERT} scrolldir = HORIZ;

enum {EDGE_TOP, EDGE_BOTTOM, EDGE_LEFT, EDGE_RIGHT};
int hasedge[4] = {1, 1, 1, 1};

#define C_BLACK 'B'
#define C_WHITE 'W'
#define C_EMPTY '.'

char board[19 * 19];

// Windows
TWindow *igo_mainmenu_win;
TWindow *igo_gamemenu_win;
TWindow *igo_mainwin;

coord_t screen_size;
coord_t mainmenu_tl;
coord_t mainmenu_size;
coord_t gamemenu_tl;
coord_t gamemenu_size;

// Menus and handler prototypes
TWindow *igo_mainmenu_quit(ttk_menu_item *item);
TWindow *igo_mainmenu_startgame(ttk_menu_item *item);
void igo_mainmenu_boardsize(ttk_menu_item *item, int cdata);
void igo_mainmenu_computer(ttk_menu_item *item, int cdata);
void igo_mainmenu_komi(ttk_menu_item *item, int cdata);
void igo_mainmenu_handicap(ttk_menu_item *item, int cdata);

TWindow *igo_gamemenu_return(ttk_menu_item *item);
TWindow *igo_gamemenu_pass(ttk_menu_item *item);
TWindow *igo_gamemenu_quit(ttk_menu_item *item);

const char *bsize_choices[] = {"9x9", "13x13", NULL};
const char *cmp_choices[] = {"Black", "White", NULL};
const char *handi_choices[] = {"0", "2", "3", "4", "5", "6", "7", "8", "9", NULL};
const char *komi_choices[] = {"0", "0.5", "5.5", "6.5", "7.5", NULL};

ttk_menu_item mainmenu_items[] = {
    {"Start Game", {igo_mainmenu_startgame}},
    {"Board Size", .choicechanged = igo_mainmenu_boardsize, .choices = bsize_choices},
    {"Computer", .choicechanged = igo_mainmenu_computer, .choices = cmp_choices},
    {"Handicap", .choicechanged = igo_mainmenu_handicap, .choices = handi_choices},
    {"Komi", .choicechanged = igo_mainmenu_komi, .choices = komi_choices},
    {"Quit iGo", {igo_mainmenu_quit}},
    {NULL}
};

ttk_menu_item gamemenu_items[] = {
    {"Return to game", {igo_gamemenu_return}},
    {"Pass", {igo_gamemenu_pass}},
    {"Quit Game", {igo_gamemenu_quit}},
    {NULL}
};

// Helper functions for coordinates
coord_t coord_add(coord_t pos, int x, int y)
{
    pos.x += x;
    pos.y += y;
    return pos;
}

coord_t coord(int x, int y)
{
    coord_t ret;
    ret.x = x;
    ret.y = y;
    return ret;
}

int in_board(coord_t pos) {
    return (pos.x >= 0 && pos.x < board_size &&
	    pos.y >= 0 && pos.y < board_size);
}

char *colname(int c)
{
    static char *cols[] = {"black", "white"};
    return cols[(c == C_BLACK) ? 0 : 1];
}

void gnugo_send(char *line)
{
    fprintf(stderr, "send: %s\n", line);
    fprintf(gnugo_stdin, "%s\n", line);
    fflush(gnugo_stdin);
}

void board_set(coord_t pos, int c)
{
    if (in_board(pos)) {
	board[pos.y * board_size + pos.x] = c;
    }
}

int board_get(coord_t pos)
{
    if (in_board(pos)){
	return board[pos.y * board_size + pos.x];
    } else {
	return -1;
    }
}

// Place handicap stones to board
void place_handi(int h)
{
    int i;
    int s_9[9][2] = {{6, 2}, {2, 6}, {2, 2}, {6, 6},
		     {2, 4}, {6, 4}, {4, 2}, {4, 6}, {4, 4}};
    int s_13[9][2] = {{9, 3}, {3, 9}, {3, 3}, {9, 9},
		      {3, 6}, {9, 6}, {6, 3}, {6, 9}, {6, 6}};

    if (h == 5 || h == 7) {
	if (board_size == 9) {
	    board_set(coord(4, 4), C_BLACK);
	} else {
	    board_set(coord(6, 6), C_BLACK);
	}
	h--;
    }

    for (i = 0; i < h; i++) {
	if (board_size == 9) {
	    board_set(coord(s_9[i][0], s_9[i][1]), C_BLACK);
	} else {
	    board_set(coord(s_13[i][0], s_13[i][1]), C_BLACK);
	}
    }
}

int has_libs(coord_t pos, int c, char *m)
{
    fprintf(stderr, "has_libs %d %d\n", pos.x, pos.y);

    if (!in_board(pos)) {
	fprintf(stderr, "out of board at %d %d\n", pos.x, pos.y);
	return 0;
    } else if (m[pos.y * board_size + pos.x]) {
	fprintf(stderr, "marked visited %d %d\n", pos.x, pos.y);
	return 0;
    } else if (board_get(pos) == C_EMPTY) {
	fprintf(stderr, "lib at %d %d\n", pos.x, pos.y);
	return 1;
    } else if (board_get(pos) != c) {
	fprintf(stderr, "opponent stone at %d %d\n", pos.x, pos.y);
	return 0;
    } else {
	int libs;
	m[pos.y * board_size + pos.x] = 1;

	libs = (has_libs(coord_add(pos, 1, 0), c, m) +
		has_libs(coord_add(pos, -1, 0), c, m) +
		has_libs(coord_add(pos, 0, 1), c, m) +
		has_libs(coord_add(pos, 0, -1), c, m));
	if (!libs) {
	    fprintf(stderr, "no libs %d %d\n", pos.x, pos.y);
	    return 0;
	} else {
	    fprintf(stderr, "has libs %d %d\n", pos.x, pos.y);
	    return 1;
	}
    }
}

// Recursively capture group of color c at x,y, return number of captured stones
int capture_group(coord_t pos, int c)
{
    fprintf(stderr, "capture group %d %d\n", pos.x, pos.y);

    if (in_board(pos) && board_get(pos) == c) {
	board_set(pos, C_EMPTY);
	return (capture_group(coord_add(pos, 1, 0), c) +
		capture_group(coord_add(pos, -1, 0), c) +
		capture_group(coord_add(pos, 0, 1), c) +
		capture_group(coord_add(pos, 0, -1), c) + 1);
    } else {
	return 0;
    }
}

// Check whether group of color c at pos should be captured.
// If it should, remove stones from board and return number of captured
// stones.
int try_capture(coord_t pos, int c, char *m)
{
    fprintf(stderr, "try capture %d %d\n", pos.x, pos.y);
    
    if (in_board(pos) && (board_get(pos) == c) &&
	!m[pos.y * board_size + pos.x] && !has_libs(pos, c, m)){
	return capture_group(pos, c);
    } else {
	return 0;
    }
}

// When a stone is played, all adjacted groups are candidates for capture
int do_captures(coord_t pos, int c)
{
    int cs = 0;
    static char m[19 * 19]; // enough for largest board size
    memset(m, 0, 19 * 19);
    m[pos.y * board_size + pos.x] = 1;

    cs += try_capture(coord_add(pos, 1, 0), c, m);
    cs += try_capture(coord_add(pos, -1, 0), c, m);
    cs += try_capture(coord_add(pos, 0, 1), c, m);
    cs += try_capture(coord_add(pos, 0, -1), c, m);

    return cs;
}

// fill board with C_EMPTY
void clear_board()
{
    int i, j;

    for (i = 0; i < board_size; i++) {
	for (j = 0; j < board_size; j++) {
	    board_set(coord(i, j), C_EMPTY);
	}
    }
}

// print free memory
void draw_meminfo(ttk_surface s, int x, int y)
{
    char unknown[] = "free mem: ???";
    char str[64];
    char buf[64];
    int freebytes = -1;
    FILE *f = fopen("/proc/meminfo", "rt");

    if (!f) {
	strcpy(str, unknown);
    } else {
	while (fgets(buf, 64, f)) {
	    char tok[32];
	    int p = 0;
	    if ((p = token(buf, ' ', p, 32, tok, NULL)) > 0) {
		if (!strncmp(tok, "MemFree", 7)) {
		    if (token(buf, ' ', p, 32, tok, NULL) > 0) {
			sscanf(tok, "%d", &freebytes);
		    }
		}
	    }
	}
	fclose(f);
    }

    snprintf(str, 64, "free mem: %d", freebytes);
    ttk_fillrect(s, x, y, 176, 132, col_black);
    ttk_text(s, font, x, y, col_white, str);
}

// Draw tile to board coordinates bx, by
void draw_tile(ttk_surface s, coord_t bpos, int cursor)
{
    int tx = bpos.x * tilesize + 5;
    int ty = bpos.y * tilesize + 5;
    int tilenum = T_EMPTY;

    switch (board_get(bpos)) {
    case C_BLACK:
	tilenum = T_BLACK;
	break;
    case C_WHITE:
	tilenum = T_WHITE;
	break;
    case C_EMPTY:
    {
	// edges & corners
	int le = (bpos.x == 0 && hasedge[EDGE_LEFT]);
	int re = (bpos.x == (board_size - 1) && hasedge[EDGE_RIGHT]);
	int te = (bpos.y == 0 && hasedge[EDGE_TOP]);
	int be = (bpos.y == (board_size - 1) && hasedge[EDGE_BOTTOM]);

	if (te) {
	    if (le) {
		tilenum = T_TL;
	    } else if (re) {
		tilenum = T_TR;
	    } else {
		tilenum = T_TOP;
	    }	    
	} else if (be) {
	    if (le) {
		tilenum = T_BL;
	    } else if (re) {
		tilenum = T_BR;
	    } else {
		tilenum = T_BOTTOM;
	    }
	} else if (re) {
	    tilenum = T_RIGHT;
	} else if (le) {
	    tilenum = T_LEFT;
	} else {
	    tilenum = T_EMPTY;
	}
	break;
    }
    }

    switch (board_size) {
    case 9:
	ttk_blit_image_ex(tiles_9, tilenum * tilesize, 0, tilesize, tilesize, s, tx, ty);
	break;
    case 13:
	ttk_blit_image_ex(tiles_13, tilenum * tilesize, 0, tilesize, tilesize, s, tx, ty);
	break;
    }

    if (cursor) {
	ttk_rect(s, tx, ty, tx + tilesize, ty + tilesize, col_green);
    }
}

// Draw something to entertain the player while gnugo is thinking
void draw_wait(PzWidget *w, ttk_surface s)
{
    static int n = 0;

    if (++n > 7) { 
	n = 0;
    }
    
    ttk_fillrect(backbuf, 168, 0, 176, 8, col_black);
    ttk_line(backbuf, 168, n, 176, n, col_green);
    ttk_blit_image(backbuf, s, 0, 0);
}

// Draw turn number and captured stones
void draw_gameinfo(ttk_surface s)
{
    char buf[32];
    ttk_fillrect(backbuf, 117, 0, 176, 132, col_black);
    snprintf(buf, 32, "Turn %d", movenum);
    ttk_text(s, font, 117, 10, col_white, buf);
    snprintf(buf, 32, "B: %d", b_captures);
    ttk_text(s, font, 117, 30, col_white, buf);
    snprintf(buf, 32, "W: %d", w_captures);
    ttk_text(s, font, 117, 50, col_white, buf);
    snprintf(buf, 32, "Komi: %.1f", komi);
    ttk_text(s, font, 117, 70, col_white, buf);
}

// Draw board etc during game
void draw_game(PzWidget *w, ttk_surface s)
{
    int x, y;

    ttk_fillrect(s, 0, 0, 176, 132, col_black);
    if (redraw == RD_FULL) {
	for (y = 0; y < board_size; y++) {
	    for (x = 0; x < board_size; x++) {
		draw_tile(backbuf, coord(x, y), (x == cursor_pos.x && y == cursor_pos.y) ? 1 : 0);
	    }
	}
    } else {
	draw_tile(backbuf, cpos_old, 0);
	draw_tile(backbuf, cursor_pos, 1);
    }
    draw_gameinfo(backbuf);
    //draw_meminfo(backbuf, 5, 118);
    ttk_blit_image(backbuf, s, 0, 0);
}

// 
void draw_igo(PzWidget *w, ttk_surface s)
{
    switch (gamestate) {

    case WHITE_WAIT:
    case BLACK_WAIT:
	draw_game(w, s);
	draw_wait(w, s);
	break;

    case WHITE_TURN:
    case BLACK_TURN:
    case GAMEOVER:
	draw_game(w, s);
	break;

    default:
	break;
    }
}

int human_play(coord_t pos, int c)
{
    char buf[64];
    int xchar = (char)('A' + pos.x);
    int ycoord = board_size - pos.y;
    struct pollfd ufds;

    if (pos.x > 7) { // No I
	xchar += 1;
    }

    if (c == C_BLACK) {
	b_last = pos;
    } else {
	w_last = pos;
    }

    // send to gnugo
    ufds.fd = fileno(gnugo_stdin);
    ufds.events = POLLOUT;
    //if (poll(&ufds, 1, 1000)) {
    if (1) {
	if (pos.x >= 0 && pos.y >= 0) {
	    snprintf(buf, 64, "play %s %c%d\n", colname(c), xchar, ycoord);
	} else {
	    snprintf(buf, 64, "play %s pass\n", colname(c));
	}
	fprintf(stderr, "%s\n", buf);

	// cmd to gnugo
	gnugo_send(buf);

	return 1;
    } else {
	return 0;
    }
}

// Tell gnugo to generate a move
int gnugo_play(int c)
{
    char buf[32];
    struct pollfd ufds;

    snprintf(buf, 32, "genmove_%s", colname(c));

    ufds.fd = fileno(gnugo_stdin); // to gnugo stdin
    ufds.events = POLLOUT;
//    if (poll(&ufds, 1, 250)) {
    if (1) {
	gnugo_send(buf);
	gamestate = (c == C_WHITE) ? WHITE_WAIT : BLACK_WAIT;
	return 1;
    } else {
	return 0;
    }
}

coord_t parse_coord(char *str) 
{
    coord_t c = coord(-1, -1);
    int x, y;

    fprintf(stderr, "parse_coord %s\n", str);


    x = str[0] - 'A';
    if (x > 7) { // No I
	x--;
    }
    c.x = x;
    if (sscanf(&str[1], "%d", &y) == 1){
	c.y = board_size - y;
    }

    fprintf(stderr, "%d, %d\n", c.x, c.y);

    return c;
}

int gnugo_poll()
{
    static char buf[64];
    static int bpos = 0;
    char line[64];

    struct pollfd ufds;
    int ret = GP_NOTHING;
    
    fprintf(stderr, "gnugo_poll() state %d\n", gamestate);
    
    ufds.fd = fileno(gnugo_stdout);
    ufds.events = POLLIN;
    while (poll(&ufds, 1, 50) && !(ufds.revents & (POLLHUP | POLLERR | POLLNVAL))) {
	int i = 0;
	bpos += fread(&buf[bpos], 1, 64 - bpos, gnugo_stdout);

	while (i<bpos) {

	    // handle complete lines
	    if (buf[i] == '\n') {
		char cbuf[16];

		memcpy(line, buf, i);
		line[i] = 0;

		fprintf(stderr, "state %d line: %s\n", gamestate, line);

		switch (gamestate) {
		case WHITE_WAIT:
		    if (w_computer) {
			if (sscanf(line, "= %s", cbuf) == 1) {
			    coord_t c = parse_coord(cbuf);
			    if (in_board(c)) {
				w_last = c;
				ret = GP_MOVEOK;
			    } else if (!strncmp(cbuf, "PASS", 4)) {
				ret = GP_PASS;
			    } 
			}
		    } else {
			if (line[0] == '=') {
			    ret = GP_MOVEOK;
			} else if (line[0] == '?') {
			    ret = GP_INVALID;
			}
		    }
		    break;

		case BLACK_WAIT:
		    if (b_computer) {
			if (sscanf(line, "= %s", cbuf) == 1) {
			    coord_t c = parse_coord(cbuf);
			    if (in_board(c)) {
				b_last = c;
				ret = GP_MOVEOK;
			    } else if (!strncmp(cbuf, "PASS", 4)) {
				ret = GP_PASS;
			    } 
			}
		    } else {
			if (line[0] == '=') {
			    ret = GP_MOVEOK;
			} else if (line[0] == '?') {
			    ret = GP_INVALID;
			}
		    }
		    break;

		case SETUP:
		    if (line[0] == '=') {
			fprintf(stderr, "setup ok (line %s)\n", line);
			ret = GP_MOVEOK;
		    } else if (line[0] == '?') {
			ret = GP_INVALID;
		    }
		    break;
		    
		case COUNT_SCORE:
		{
		    char w;
		    fprintf(stderr, "%s\n", line);
		    if (!strncmp(line, "= 0", 3)) {
			winner = C_EMPTY;
			score = 0.0;
			ret = GP_MOVEOK;
		    }
		    else if (sscanf(line, "= %c+%f", &w, &score) == 2) {
			winner = (w == 'B' ? C_BLACK : C_WHITE);
			ret = GP_MOVEOK;
		    }
		    break;
		}
		    
		default:
		    break;
		}

		// move rest to buf beginning
		memmove(buf, &buf[i+1], bpos - i - 1);
		bpos -= i + 1;
		i = 0;
	    } else {
		i++;
	    }
	}
    }

    return ret;
}

// Count final score
void count_final()
{
    int ret;
    char dialog_text[32];
    fprintf(stderr, "count_final\n");

    gnugo_send("final_score");
    while ((ret = gnugo_poll()) == GP_NOTHING);
    if (winner == C_EMPTY) {
	pz_dialog("Result", "The game is a jigo.", 1, 0, "Ok");
    } else {
	snprintf(dialog_text, 32, "%s wins by %.2f",
		 (winner == C_BLACK ? "Black" : "White"), score);
	pz_dialog("Result", dialog_text, 1, 0, "Ok");
    }
}

// Handle event during game
int event_game(PzEvent *ev)
{
    switch (ev->type) {
	
    case PZ_EVENT_TIMER:

	if (gamestate == WHITE_WAIT) {

	    switch (gnugo_poll()) {
		
	    case GP_MOVEOK:
		movenum++;
		board_set(w_last, C_WHITE);
		w_captures += do_captures(w_last, C_BLACK);
		gamestate = BLACK_TURN;
		redraw = RD_FULL;
		break;

	    case GP_PASS:
		movenum++;
		w_passes++;
		if (w_passes > 0 && b_passes > 0) {
		    gamestate = COUNT_SCORE;
		    count_final();
		    gamestate = GAMEOVER;
		} else {
		    gamestate = BLACK_TURN;
		}
		break;

	    case GP_INVALID:
		gamestate = WHITE_TURN;

	    case GP_NOTHING:
		break;
	    }

	} else if (gamestate == BLACK_WAIT) {

	    switch (gnugo_poll()) {
		
	    case GP_MOVEOK:
		movenum++;
		board_set(b_last, C_BLACK);
		b_captures += do_captures(b_last, C_WHITE);
		gamestate = WHITE_TURN;
		redraw = RD_FULL;
		break;

	    case GP_PASS:
		movenum++;
		b_passes++;
		if (w_passes > 0 && b_passes > 0) {
		    gamestate = COUNT_SCORE;
		    count_final();
		    gamestate = GAMEOVER;
		} else {
		    gamestate = WHITE_TURN;
		}
		break;

	    case GP_INVALID:
		gamestate = BLACK_TURN;
		break;

	    case GP_NOTHING:
		break;
	    }

	}
	ev->wid->dirty = 1;
	break;
	
    case PZ_EVENT_SCROLL:
	cpos_old = cursor_pos;
	
	TTK_SCROLLMOD(ev->arg, 4);
	if (ev->arg > 0) {
	    if (scrolldir == HORIZ) {
		cursor_pos.x = MIN(board_size - 1, cursor_pos.x + 1);
	    } else {
		cursor_pos.y = MIN(board_size - 1, cursor_pos.y + 1);
	    }
	} else {
	    if (scrolldir == HORIZ) {
		cursor_pos.x = MAX(0, cursor_pos.x - 1);
	    } else {
		cursor_pos.y = MAX(0, cursor_pos.y - 1);
	    }
	}
	redraw = RD_CURSOR;
	ev->wid->dirty = 1;
	break;
	
    case PZ_EVENT_BUTTON_DOWN:
	
	switch (ev->arg) {
	    
	case PZ_BUTTON_PLAY:
	    // change between horiz/vert scroll
	    scrolldir = (scrolldir == HORIZ) ? VERT : HORIZ;
	    break;
	    
	case PZ_BUTTON_ACTION:
	    if (gamestate == BLACK_TURN && !b_computer) {
		if (human_play(cursor_pos, C_BLACK)) {
		    gamestate = BLACK_WAIT;
		    redraw = RD_FULL; // redraw whole board
		}
		ev->wid->dirty = 1;
	    } else if (gamestate == WHITE_TURN && !w_computer) {
		if (human_play(cursor_pos, C_WHITE)) {
		    gamestate = WHITE_WAIT;
		    redraw = RD_FULL;
		}
		ev->wid->dirty = 1;
	    }
	    break;
	}
	
    case PZ_EVENT_BUTTON_UP:
	switch (ev->arg) {
	case PZ_BUTTON_MENU:
	    ttk_show_window(igo_gamemenu_win);
	    break;
	}
	break;
	    
    }
	return 0;
}

// Handle event in menu
int event_title(PzEvent *ev)
{
    switch (ev->type) {

    case PZ_EVENT_BUTTON_DOWN:

	switch (ev->arg) {

	case PZ_BUTTON_ACTION:
	    gamestate = BLACK_TURN;
	    ev->wid->dirty = 1;
	    break;
	}
	break;

    case PZ_EVENT_TIMER:
	ev->wid->dirty = 1;
	break;
    }

    return 0;
}

// Handle event
int event_igo(PzEvent *ev)
{
    switch (gamestate) {

    case MAINMENU:
	event_title(ev);
	break;

    case WHITE_TURN:
	if (w_computer && gnugo_play(C_WHITE)) {
	    gamestate = WHITE_WAIT;
	} else {
	    event_game(ev);
	}
	break;

    case BLACK_TURN:
	if (b_computer && gnugo_play(C_BLACK)) {
	    gamestate = BLACK_WAIT;
	} else {
	    event_game(ev);
	}
	break;

    case WHITE_WAIT:
    case BLACK_WAIT:
	event_game(ev);
	break;

    default:
	break;
    }

    return 0;
}

// Free resources on module unload
void cleanup_igo(void)
{
    ttk_done_font(font);
}

// Fork a gnugo-process
int launch_gnugo()
{
    int gg_stdin, gg_stdout;

    mkfifo(gnugo_pipe_to, S_IRUSR | S_IWUSR);
    mkfifo(gnugo_pipe_from, S_IRUSR | S_IWUSR);
    if ((gg_stdout = open(gnugo_pipe_from, O_RDONLY | O_NONBLOCK)) < 0) {
	return -1;
    }

    if ((gnugo_pid = vfork()) > 0) {
	// parent
	if ((gg_stdin = open(gnugo_pipe_to, O_WRONLY)) < 0) {
	    return -1;
	}
	
	if (!(gnugo_stdin = fdopen(gg_stdin, "w")) ||
	    !(gnugo_stdout = fdopen(gg_stdout, "r"))) {
	    return -1;
	}

    } else if (gnugo_pid == 0) {
	int outdesc;

	if ((outdesc = open(gnugo_pipe_from, O_WRONLY)) < 0) {
	    return -1;
	}
	dup2(outdesc, 1); // gnugo stdout should go to pipe

	if (board_size == 9) {
	    execl(GNUGO_COMMAND, GNUGO_ARGS_9);
	} else {
	    execl(GNUGO_COMMAND, GNUGO_ARGS_13);
	}
	return -1;
    } else {
	return -1;
    }
    return 0;
}


//
void kill_gnugo()
{
    if (gnugo_pid > 0) {
	int status;
	kill(gnugo_pid, SIGTERM);
	waitpid(gnugo_pid, &status, 0);
	gnugo_pid = 0;
    }
}

// reset game state
void init_game()
{
    clear_board();
    b_passes = w_passes = 0;
    b_captures = w_captures = 0;
    winner = -1;
    score = 0.0;
    movenum = 1;
}

// Construct main menu
PzWindow *make_igo_mainmenu()
{
    TWindow *win;
    TWidget *menu_widget = ttk_new_menu_widget(mainmenu_items, font,
					       mainmenu_size.x,
					       mainmenu_size.y);
    win = ttk_new_window();
    win->x = mainmenu_tl.x;
    win->y = mainmenu_tl.y;
    win->w = mainmenu_size.x;
    win->h = mainmenu_size.y;
    ttk_add_widget(win, menu_widget);
    ttk_window_set_title(win, "iGo 0.1");
    return win;
}


// Construct game menu
PzWindow *make_igo_gamemenu()
{
    TWindow *win;
    TWidget *menu_widget = ttk_new_menu_widget(gamemenu_items, font,
					       gamemenu_size.x,
					       gamemenu_size.y);
    win = ttk_new_window();
    win->x = gamemenu_tl.x;
    win->y = gamemenu_tl.y;
    win->w = gamemenu_size.x;
    win->h = gamemenu_size.y;
    ttk_add_widget(win, menu_widget);
    ttk_window_hide_header(win);
    ttk_set_popup(win);
    return win;
}

// Construct main game window
PzWindow *make_igo_mainwin()
{
    PzWindow *win;

    win = pz_new_window("iGo", PZ_WINDOW_NORMAL);
    ttk_window_hide_header(win);

    widget = pz_add_widget(win, draw_igo, event_igo );
    pz_widget_set_timer(widget, 100);

    return win;
}

// Menu handlers
TWindow *igo_mainmenu_quit(ttk_menu_item *item)
{
    return TTK_MENU_UPONE;
}

TWindow *igo_mainmenu_startgame(ttk_menu_item *item)
{
    int gnugo_err;
    char cmd[32];
    init_game();
    if ((gnugo_err = launch_gnugo()) != 0) {
	pz_errdialog ("terror", "gnugo didn't start", 1, 100, "ok");
    }

    // set komi & handi
    gamestate = SETUP;
    snprintf(cmd, 32, "komi %.1f", komi);
    gnugo_send(cmd);
    if (!gnugo_err) {
	while (gnugo_poll() == GP_NOTHING);
    }
    if (handi > 0) {
	snprintf(cmd, 32, "fixed_handicap %d", handi);
	gnugo_send(cmd);
	while (gnugo_poll() == GP_NOTHING);
    }

    place_handi(handi);
    gamestate = (handi == 0) ? BLACK_TURN : WHITE_TURN;

    // There's no need to reconstruct the main window, but this handler
    // should be called every time the item is selected.
    redraw = RD_FULL;
    ttk_show_window(igo_mainwin);
    return TTK_MENU_DONOTHING;
}

void igo_mainmenu_boardsize(ttk_menu_item *item, int cdata)
{
    switch (item->choice) {
    case 0:
	board_size = 9;
	tilesize = 12;
	break;
    case 1:
	board_size = 13;
	tilesize = 8;
	break;
    }
}

void igo_mainmenu_computer(ttk_menu_item *item, int cdata)
{
    if (item->choice == 0) {
	b_computer = !(w_computer = 0);
    } else {
	w_computer = !(b_computer = 0);
    }
}

void igo_mainmenu_handicap(ttk_menu_item *item, int cdata)
{
    if (item->choice == 0) {
	handi = 0;
    } else {
	handi = item->choice + 1;
    }
}

void igo_mainmenu_komi(ttk_menu_item *item, int cdata)
{
    komi = atof(komi_choices[item->choice]);
}

TWindow *igo_gamemenu_return(ttk_menu_item *item)
{
    return TTK_MENU_UPONE;
}

TWindow *igo_gamemenu_pass(ttk_menu_item *item)
{
    switch (gamestate) {
    case BLACK_TURN:
	b_passes++;
	human_play(coord(-1, -1), C_BLACK);
	gamestate = BLACK_WAIT;
	break;

    case WHITE_TURN:
	w_passes++;
	human_play(coord(-1, -1), C_WHITE);
	gamestate = WHITE_WAIT;
	break;

    default:
	break;
    }

    return TTK_MENU_UPONE;
}

TWindow *igo_gamemenu_quit(ttk_menu_item *item)
{
    kill_gnugo();
    init_game();
    gamestate = MAINMENU;
    ttk_hide_window(igo_gamemenu_win);
    ttk_hide_window(igo_mainwin);
    
    return TTK_MENU_DONOTHING;
}

void init_igo()
{
    module = pz_register_module ("igo", cleanup_igo);
    pz_menu_add_action ("/Extras/Games/iGo", make_igo_mainmenu);
    
    tiles_9 = ttk_load_image(pz_module_get_datapath(module, "tiles.png"));
    tiles_13 = ttk_load_image(pz_module_get_datapath(module, "tiles13.png"));
    font = ttk_get_font("Fixed 6x13", 13);

    backbuf = ttk_new_surface(176, 132, 16);
    ttk_fillrect(backbuf, 0, 0, 176, 132, col_black);

    screen_size = coord(ttk_screen->w, ttk_screen->h);
    mainmenu_tl = coord(ttk_screen->wx, ttk_screen->wy);
    mainmenu_size = coord(ttk_screen->w - ttk_screen->wx,
			  ttk_screen->h - ttk_screen->wy);
    gamemenu_tl = coord(10, 10);
    gamemenu_size = coord(100, 100);

    igo_mainwin = make_igo_mainwin();
    igo_mainmenu_win = make_igo_mainmenu();
    igo_gamemenu_win = make_igo_gamemenu();

    col_black = ttk_makecol(0, 0, 0);
    col_white = ttk_makecol(255, 255, 255);
    col_green = ttk_makecol(0, 255, 0);

    cursor_pos = coord(0, 0);
    cpos_old = coord(0, 0);
}

PZ_MOD_INIT(init_igo)
